import asyncio
import logging
from typing import Optional

# https://svn.nmap.org/nmap/nmap-service-probes

class ProtocolError(Exception):
    pass

class UnsupportedVersion(Exception):
    pass

class UdpTransporter():
    def __init__(self, dst_ip = "0.0.0.0", dst_port=1, src_ip="0.0.0.0", src_port=1):
        self.dst_ip = dst_ip
        self.dst_port = dst_port
        self.src_ip = src_ip
        self.src_port = src_port

    def get_extra_info(self, name, default=None):
        #https://docs.python.org/3/library/asyncio-protocol.html
        if name == 'sockname':
            return self.dst_ip, self.dst_port
        elif name == 'peername':
            return self.src_ip, self.src_port
        else:
            return None

class BaseProtocol(asyncio.Protocol):
    """common class to all protocol handler"""
    def __init__(self):
        self.protocol_name = "base"

    def connection_made(self, transport) -> None:
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)
        self.transport.close()
        
    def data_received(self, data):
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
        self.transport.close()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self.transport.close()

    def unrecognized_data(self, data):
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
        self.transport.close()

class BaseHoneypot(object):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0", **kwargs):
        self.port = config['port']
        self.bindaddr = bindaddr
        self.logger = logger 
        self.handler = BaseProtocol
        self.handler.logger = logger
        self.server = None
        self.task = None    

    async def start(self):
        # Start the server in a separate task
        loop = asyncio.get_running_loop()
        self.task = loop.create_task(self._start_server())
        return self.task
    
    async def _start_server(self):
        loop = asyncio.get_running_loop()
        try:
            self.server = await loop.create_server(self.handler, host=self.bindaddr, port=self.port)
            await self.server.serve_forever()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.error(e)
            return False

    async def stop(self):
        self.task.cancel()
        try:
            await self.task
        except asyncio.CancelledError:
            logging.info(f"Server {self.bindaddr}:{self.port} is cancelled now")



from trapster.modules.base import BaseProtocol, BaseHoneypot, UdpTransporter
from trapster.libs import dns

import asyncio, logging
        
class EchoClientProtocol(asyncio.DatagramProtocol):
    def __init__(self, message, on_con_lost):
        self.message = message
        self.on_con_lost = on_con_lost
        self.transport = None
        self.return_data = b''

    def connection_made(self, transport):
        self.transport = transport
        self.transport.sendto(self.message)

    def datagram_received(self, data, addr):        
        self.return_data = data
        self.transport.close()

    def connection_lost(self, exc):
        try:
            self.on_con_lost.set_result(True)
        except asyncio.exceptions.InvalidStateError:
            pass


class DnsUdpProtocol(BaseProtocol):

    config = {
        'dns1': "127.0.0.1"
    }

    def __init__(self, config=None):
        self.protocol_name = "dns"

        if config:
            self.config = config

    def connection_made(self, transport) -> None:
        self.transport = transport
        
    def datagram_received(self, data, addr):
        # decode dns packet to json
        decoded_packet = dns.decode_dns_message(data)    

        # proxy packet to legit dns server
        self.loop = asyncio.get_running_loop()
        self.loop.create_task(self.proxy_packet(data, addr, decoded_packet))

    async def proxy_packet(self, data, addr, decoded_packet):
        on_con_lost = self.loop.create_future()

        # need to specify src_ip and src_port because self.transport endpoint is not connected
        src_ip, src_port = addr
        dst_ip, dst_port = self.transport.get_extra_info('sockname')
        transport_udp = UdpTransporter(dst_ip, dst_port, src_ip, src_port)
        self.logger.log(self.protocol_name + "." + self.logger.QUERY, transport_udp, extra={"query": decoded_packet})


        transport, protocol = await self.loop.create_datagram_endpoint(
            lambda: EchoClientProtocol(data, on_con_lost),
            remote_addr=(self.config['dns1'], 53))

        try:
            # send back data from the legit dns server
            await on_con_lost
            self.transport.sendto(protocol.return_data, addr)
        finally:
            transport.close()        


class DnsTcpProtocol(BaseProtocol):
    """
    For now it indicates that DNS is enable to attackers scanning only TCP
    """
    #TODO

    def __init__(self):
        self.protocol_name = "dns"  

class DnsHoneypot(BaseHoneypot):

    def __init__(self, config, logger, bindaddr, proxy_dns_ip):
        super().__init__(config, logger, bindaddr)
        # binaddr 0.0.0.0 is not accepted because real dns server is running on 127.0.0.1
        # so it mused be set in the conf file
        self.handler = lambda: DnsTcpProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config

        self.handler_udp = DnsUdpProtocol
        self.handler_udp.config['dns1'] = proxy_dns_ip

    async def _start_server(self):
        loop = asyncio.get_running_loop()

        # Create UDP server
        transport, protocol = await loop.create_datagram_endpoint(lambda: self.handler_udp(), 
                                    local_addr=(self.bindaddr, self.port))
        
        # Create TCP server
        try:
            self.server = await loop.create_server(self.handler, host=self.bindaddr, port=self.port)
            await self.server.serve_forever()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            logging.error(e)
            return False



from trapster.modules.base import BaseProtocol, BaseHoneypot
from trapster.libs import ldapasn1

from pyasn1.codec.ber import decoder

from datetime import datetime, timezone


class LdapProtocol(BaseProtocol):

    config = {
        "server" : "server-01",
        "domain" : "microsoft",
        "tld" : "intra",
        "level": "WinThreshold"
    }

    def __init__(self, config=None):
        self.protocol_name = "ldap"

        if config:
            self.config = config
        
        self.functionality_level = self.get_functionality_level(self.config.get('level', 'WinThreshold'))
        
    def connection_made(self, transport) -> None:
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

    def data_received(self, data):
        # process request
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
        
        responses = self.process_request(data)
        if responses:
            for response in responses:
                self.send_response(response)    

    def process_request(self, request):
        # parse LDAP message using impacket ldapasn1
        ldapMessage, _ = decoder.decode(request, asn1Spec=ldapasn1.LDAPMessage())

        # dispatch message
        (response_ops, message_id) = self.dispatch(ldapMessage)
        ldapResponses=[]

        if response_ops:
                for response_op in response_ops:
                        ldapResponse = ldapasn1.LDAPMessage()
                        ldapResponse['messageID'] = message_id
                        ldapResponse['protocolOp'].setComponentByType(response_op.getTagSet(), response_op)
                        ldapResponses.append(ldapResponse)
                return ldapResponses
        else:
                return None


    def dispatch(self, ldapMessage):
        message_id = ldapMessage['messageID']
        protocolOp = ldapMessage['protocolOp']
        name = protocolOp.getName()

        msgs=[]

        if name == 'bindRequest':
            msgs.append(self.bind_response(protocolOp))
        elif name == 'searchRequest':
            msgs.append(self.searchrequest_response(protocolOp))
            msgs.append(self.searchresult_done())
        elif name == 'unbindRequest':
            self.transport.close()
        elif name == 'abandonRequest':
            self.transport.close()
        else:
            self.transport.close()

        return (msgs, message_id)


    def bind_response(self, protocolOp):
        
        authentication = protocolOp['bindRequest']['authentication'].getName()
        msg = ldapasn1.BindResponse()

        if authentication == 'simple':
            username = str(protocolOp['bindRequest']['name'])
            password = str(protocolOp['bindRequest']['authentication']['simple'])

            if password == '':
                # anonymous bind, allow
                msg['resultCode'] = 0
                msg['matchedDN'] = ''
                msg['diagnosticMessage'] = ''
            else:
                # register user, disallow
                msg['resultCode'] = 49
                msg['matchedDN'] = ''
                msg['diagnosticMessage'] = '8009030C: LdapErr: DSID-0C090569, comment: AcceptSecurityContext error, data 2030, v4f7c'
            
            self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={'username':username, 'password':password})

        elif authentication == 'sasl':
            #TODO
            msg['resultCode'] = 49
            msg['matchedDN'] = ''
            msg['diagnosticMessage'] = '8009030C: LdapErr: DSID-0C090569, comment: AcceptSecurityContext error, data 2030, v4f7c'
            pass

        return msg

    def searchrequest_response(self, protocolOp):
        scope = str(protocolOp['searchRequest']['scope'])
        msg = ldapasn1.SearchResultDone()

        try:
            base_object = protocolOp['searchRequest']['baseObject']
        except KeyError:
            base_object = ""

        self.logger.log(self.protocol_name + "." + self.logger.QUERY, self.transport, extra={
            'scope': scope, 
            'baseObject': str(base_object)
        })

        if scope == 'baseObject':
            # return informations about server
            # client can query many attributes without being authenticated
            # example : ldapsearch -H ldap://<ip> -s base -x serverName dnsHostName

            attributes = protocolOp['searchRequest']['attributes']
            attributes_search_values = []
            for attr in attributes:
                attributes_search_values.append(attr._value.decode(errors='backslashreplace'))

            msg = self.searchresentry_response(attributes_search_values)

        else:
            msg['resultCode'] = 1
            msg['matchedDN'] = ''
            msg['diagnosticMessage'] = 'errorMessage: 000004DC: LdapErr: DSID-0C090C21, comment: In order to perform this operation a successful bind must be completed on the connection., data 0, v4f7c'
        
        return msg

    def get_functionality_level(self, value: str):
        # Functionality level from : https://learn.microsoft.com/fr-fr/powershell/module/activedirectory/set-addomainmode?view=windowsserver2022-ps

        identifier = [
            'Windows2000Domain',
            'Windows2003InterimDomain',
            'Windows2003Domain',
            'Windows2008Domain',
            'Windows2008R2Domain',
            'Windows2012Domain',
            'Windows2012R2Domain',
            'WinThreshold', # Windows Server 2016
        ]
        try:
            return str(identifier.index(value))
        except ValueError:
            return str(7) # default to 2016

    def searchresentry_response(self, attributes_search_values):
        # default attribute list on Windows Server
        attributes_values = {
            'domainFunctionality': self.functionality_level,
            'forestFunctionality': self.functionality_level,
            'domainControllerFunctionality': self.functionality_level,
            'rootDomainNamingContext': f"DC={self.config['domain']},DC={self.config['tld']}",
            'ldapServiceName': f"{self.config['domain']}.{self.config['tld']}:{self.config['server']}$@{self.config['domain'].upper()}.{self.config['tld'].upper()}",
            'isGlobalCatalogReady': 'TRUE',
            'supportedSASLMechanisms': [
                'GSSAPI', 'GSS-SPNEGO', 'EXTERNAL', 'DIGEST-MD5'
            ],
            'supportedLDAPVersion': [
              '3', '2'
            ],
            'supportedLDAPPolicies': [
                'MaxPoolThreads','MaxPercentDirSyncRequests','MaxDatagramRecv','MaxReceiveBuffer','InitRecvTimeout','MaxConnections','MaxConnIdleTime','MaxPageSize','MaxBatchReturnMessages','MaxQueryDuration','MaxDirSyncDuration','MaxTempTableSize','MaxResultSetSize','MinResultSets','MaxResultSetsPerConn','MaxNotificationPerConn','MaxValRange','MaxValRangeTransitive','ThreadMemoryLimit','SystemMemoryLimitPercent'
            ],
            'supportedControl': [
                '1.2.840.113556.1.4.319','1.2.840.113556.1.4.801','1.2.840.113556.1.4.473','1.2.840.113556.1.4.528','1.2.840.113556.1.4.417','1.2.840.113556.1.4.619','1.2.840.113556.1.4.841','1.2.840.113556.1.4.529','1.2.840.113556.1.4.805','1.2.840.113556.1.4.521','1.2.840.113556.1.4.970','1.2.840.113556.1.4.1338','1.2.840.113556.1.4.474','1.2.840.113556.1.4.1339','1.2.840.113556.1.4.1340','1.2.840.113556.1.4.1413','2.16.840.1.113730.3.4.9','2.16.840.1.113730.3.4.10','1.2.840.113556.1.4.1504','1.2.840.113556.1.4.1852','1.2.840.113556.1.4.802','1.2.840.113556.1.4.1907','1.2.840.113556.1.4.1948','1.2.840.113556.1.4.1974','1.2.840.113556.1.4.1341','1.2.840.113556.1.4.2026','1.2.840.113556.1.4.2064','1.2.840.113556.1.4.2065','1.2.840.113556.1.4.2066','1.2.840.113556.1.4.2090','1.2.840.113556.1.4.2205','1.2.840.113556.1.4.2204','1.2.840.113556.1.4.2206','1.2.840.113556.1.4.2211','1.2.840.113556.1.4.2239','1.2.840.113556.1.4.2255','1.2.840.113556.1.4.2256','1.2.840.113556.1.4.2309','1.2.840.113556.1.4.2330','1.2.840.113556.1.4.2354'
            ],
            'supportedCapabilities' : [
                '1.2.840.113556.1.4.800','1.2.840.113556.1.4.1670','1.2.840.113556.1.4.1791','1.2.840.113556.1.4.1935','1.2.840.113556.1.4.2080','1.2.840.113556.1.4.2237' 
            ],
            'subschemaSubentry' : f"CN=Aggregate,CN=Schema,CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
            'serverName' : f"CN={self.config['server']},CN=Servers,CN=Default-First-Site-Name,CN=Sites,CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
            'schemaNamingContext' : f"CN=Schema,CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
            'namingContexts' : [
                f"DC={self.config['domain']},DC={self.config['tld']}",
                f"CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
                f"CN=Schema,CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
                f"DC=DomainDnsZones,DC={self.config['domain']},DC={self.config['tld']}",
                f"DC=ForestDnsZones,DC={self.config['domain']},DC={self.config['tld']}",
            ],
            'isSynchronized': 'TRUE',
            'highestCommittedUSN': '49175',
            'dsServiceName': f"CN=NTDS Settings,CN={self.config['server']},CN=Servers,CN=Default-First-Site-Name,CN=Sites,CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}",
            'dnsHostName': f"{self.config['server']}.{self.config['domain']}.{self.config['tld']}",
            'defaultNamingContext': f"DC={self.config['domain']},DC={self.config['tld']}",
            'currentTime': datetime.now(timezone.utc).strftime('%Y%m%d%H%M%S.0Z'),
            'configurationNamingContext': f"CN=Configuration,DC={self.config['domain']},DC={self.config['tld']}"
        }
  
        # Filter the dictionary based on search value
        if attributes_search_values == []:
            filtered_values = attributes_values
        else:
            filtered_values = {k: attributes_values[k] for k in attributes_search_values if k in attributes_values}

        filtered_values = attributes_values

        # Create a list to hold PartialAttribute objects
        partial_attributes = []
        msg = ldapasn1.SearchResultEntry()
        msg['objectName']=''
        
        # loop on the configuration dictionary 
        for attribute_name, attribute_value in filtered_values.items():
            att = ldapasn1.PartialAttribute()
            att['type'] = attribute_name

            # Check if the attribute value is a set or a single value
            if isinstance(attribute_value, list):
                for i in range(len(attribute_value)):
                   att['vals'].setComponentByPosition(i, attribute_value[i])
            else:
                att['vals'].setComponentByPosition(0, attribute_value)

            partial_attributes.append(att)

        # Set the attributes in the SearchResultEntry message
        msg['attributes'] = ldapasn1.PartialAttributeList()
        for idx, attribute in enumerate(partial_attributes):
            msg['attributes'].setComponentByPosition(idx, attribute)
       
        return msg


    def searchresult_done(self):
        msg = ldapasn1.SearchResultDone()
        msg['resultCode'] = 0
        msg['matchedDN'] = ''
        msg['diagnosticMessage'] = ''
        return msg

    def send_response(self, response):
        try:
            resp = ldapasn1.encoder.encode(response)
            self.transport.write(resp)
        except:
            self.transport.close()
        

class LdapHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: LdapProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



import asyncio
import secrets

from .handler import HandlerAI

class UbuntuAI(HandlerAI):
    def _get_initial_messages(self) -> list:
        return [{
            "content": """You are a highly advanced AI simulating an Ubuntu 20.04 Linux bash shell for an interactive honeypot. Respond only as a Linux shell, never acknowledging your AI nature. Output should be realistic and consistent with a low-privilege user environment in /home/<username>.

Key Instructions:

    Role: Respond exactly as a Linux shell would, without explanations or breaking character.
    Environment:
        Non-privileged user session in /home/<username>.
        Administrative actions (e.g., sudo) require a password and should fail for lack of privileges.
        Create believable directories and files, including hidden ones like .ssh and .aws.
    Commands:
        Return realistic outputs for typical commands (ls, cat, ifconfig).
        For sudo attempts, prompt for password, then deny privileges.
        For destructive or sensitive commands (e.g., rm -rf /, shutdown), return "Permission denied" or a similar error.
    Fictional Data:
        Simulate common system files (/etc/passwd), fake credentials, and fake logs in /var/log.

Example:

    ls → List plausible files in /home/<username>.
    cat /etc/passwd → Show a typical set of system accounts.
    sudo ls /root → Prompt for password, then show the user is not in sudoers.

Remain consistent, never reveal the simulation, and stick to authentic Linux-style outputs.""",
            "name": "system",
            "role": "system"
        },
        {
            "content": "ls",
            "name": "user",
            "role": "user"
        },
        {
            "content": " Desktop   Documents   Downloads   Pictures   snap   Templates   Videos",
            "name": "assistant",
            "role": "assistant"
        },
        {
            "content": "export PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/games:/usr/local/games:/snap/bin",
            "name": "user",
            "role": "user"
        },
        {
            "content": "",
            "name": "assistant",
            "role": "assistant"
        }]

if __name__ == "__main__":
    async def main():
        ssh_ai = UbuntuAI()
        identifier = secrets.token(8)
        while True:
            input_ = input('guest@trapster:~$ ')
            response = await ssh_ai.make_query(identifier, input_)
            if response:
                print(response)

    asyncio.run(main())




from .handler import HandlerAI

class HttpAI(HandlerAI):
    def _get_initial_messages(self) -> list:
        return [{
            "content": """You are simulating a simple HTTP API server that responds to incoming requests in JSON format, including those attempting SQL injection. You must respond as a real server would, never revealing that you are simulated. Always return JSON-style responses (API content), and if the URL includes suspicious SQL-like input, respond with subtle, believable error messages or odd behavior typical of a vulnerable web application. Never break character or acknowledge the simulation.

Key Instructions:
    Respond with JSON output as if serving a standard PHP-based web application.
    If the request looks like an SQL injection attempt, produce plausible database errors, partial data dumps, or suspicious SQL-related error messages.
    Do not reveal internal logic or say you are a honeypot.
    Keep responses authentic and consistent, as though this is an unsecured or vulnerable server.""",
            "name": "system",
            "role": "system"
            }
        ]


from redis import asyncio as aioredis
import json
import hashlib
from typing import List, Dict

class RedisManager:
    def __init__(self, host="localhost", port=6379, history_expiration=None, cache_expiration=3600):
        self.redis = None
        self.host = host
        self.port = port
        self.history_expiration = history_expiration # Purge history after expiration
        self.cache_expiration = cache_expiration # Purge cache after expiration

    async def connect(self):
        self.redis = await aioredis.from_url(
            f"redis://{self.host}:{self.port}",
            decode_responses=True  # This will automatically decode responses to strings
        )

    async def get_history(self, session_id: str) -> List[Dict]:
        history = await self.redis.get(f"history:{session_id}")
        if history:
            return json.loads(history)
        return []

    async def add_to_history(self, session_id: str, messages: Dict):
        history = await self.get_history(session_id)
        history.append(messages)
        await self.redis.set(f"history:{session_id}", json.dumps(history), ex=self.history_expiration)

    async def get_cache(self, user_message: str) -> str:
        key = self._generate_key(user_message)
        return await self.redis.get(f"cache:{key}")

    async def set_cache(self, user_message: str, response: str):
        key = self._generate_key(user_message)
        await self.redis.set(f"cache:{key}", response, ex=self.cache_expiration)

    def _generate_key(self, message: str) -> str:
        return hashlib.sha256(message.encode()).hexdigest()


import pytest
import asyncio
import socket
from unittest.mock import MagicMock

from trapster.modules.ftp import FtpHoneypot
from trapster.logger import JsonLogger

from .base import BaseServerTest

class FTPTest(BaseServerTest):
    def connect_and_login(self, port, username="username", password="randompass"):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            # Connect to the FTP server
            sock.connect((self.bindaddr, port))  

            # Read initial response
            response = sock.recv(1024).decode('utf-8')
            self.assert_response(response, 220, f"{self.service_config['banner']}\r\n")

            # Send USER command
            sock.sendall(f'USER {username}\r\n'.encode())
            response = sock.recv(1024).decode('utf-8')
            self.assert_response(response, 331, f"User {username} OK. Password required\r\n")

            # Send PASS command
            sock.sendall(f'PASS {password}\r\n'.encode())
            response = sock.recv(1024).decode('utf-8')
            self.assert_response(response, 530, "Authentication Failed\r\n")

    def assert_response(self, response, expected_code, expected_message):
        assert response.startswith(f'{expected_code}'), f"Expected {expected_code} response, got {response[:3]}"
        assert expected_message in response, f"Expected message '{expected_message}', got '{response.strip()}'"

    def validate_logger_output(self):
        """
        Validates the logger output by checking the arguments passed to the mocked log method.
        """
        # Mock the return values of the log method
        mock_return_value = {"type": "mocked_event", "extra": {}}
        self.logger.log.return_value = mock_return_value

        # After running the test, check that log was called the expected number of times
        assert self.logger.log.call_count == 4, "Expected four log entries"

        # check that extra contains the username/password
        connection_args = self.logger.log.call_args_list[3]          
        assert connection_args[1] == {'extra': {'password': 'randompass', 'username': 'username'}}, "Expected extra data with username/password"

    async def run_test(self):
        await asyncio.to_thread(self.connect_and_login, self.service_config['port'])
        self.validate_logger_output()

@pytest.mark.asyncio
async def test_ftp_authentication_failed():
    service_config = {
        'port': 2121,
        'banner': "Microsoft FTP Service"
    }

     # Mock the logger
    mock_logger = MagicMock(JsonLogger('trapster-1'))

    # Initialize the FTP-specific test
    ftp_test = FTPTest(FtpHoneypot, service_config, mock_logger)

    # Start the server
    await ftp_test.start_server()

    try:
        # Run the FTP-specific test logic
        await ftp_test.run_test()
    finally:
        # Stop the server
        await ftp_test.stop_server()




from .base import BaseProtocol, BaseHoneypot
import os, base64

class RsyncProtocol(BaseProtocol):

    config = {
        'version': "31.0",
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "rsync"
        self.user = ""
        self.password = ""
        self.expected_auth = False
        self.modules = ["Backup", "Users"]

    def connection_made(self, transport) -> None:
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

    def data_received(self, data):
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)

        if data.startswith(b"@RSYNCD:"):
            greeting = f"@RSYNCD: {self.config['version']}"
            self.transport.write(greeting.encode() + b"\n")
            return

        elif data == b"\n":
            for module in self.modules:
                self.transport.write(f"{module}    	{module}\n".encode())
            self.transport.write(b"@RSYNCD: EXIT\n")
            return
        
        module = data.decode().strip()
        if module in self.modules:
            self.requested_module = module
            self.auth_challenge = self.generate_challenge()
            self.transport.write(f"@RSYNCD: AUTHREQD {self.auth_challenge}\n".encode())
            self.expected_auth = True
            return

        # Step 4: Handle authentication response
        elif self.expected_auth:
            auth_data = data.decode().strip().split(" ", 1)
            if len(auth_data) != 2:
                self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": auth_data})
                self.transport.write(f"@ERROR: auth failed on module {self.requested_module}\n".encode())
                return

            username, hashed_response = auth_data
            self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": str(username), "password": str(hashed_response)})

            # Always deny authentication (honeypot behavior)
            self.transport.write(f"@ERROR: auth failed on module {self.requested_module}\n".encode())
            self.expected_auth = False  # Reset authentication state
            return

        else:
            self.logger.log(self.protocol_name + "." + self.logger.QUERY, self.transport, extra={"module": data.decode('utf-8', errors='replace')})
            self.transport.write(f"@ERROR: unknown module '{module}'\n".encode())
            return

    def generate_challenge(self):
        """Generate a base64 random challenge string (16 bytes)."""
        return base64.b64encode(os.urandom(16)).decode()
    

class RsyncHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: RsyncProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



from trapster.modules.base import BaseProtocol, BaseHoneypot

class MssqlProtocol(BaseProtocol):
    def __init__(self, config=None):
        self.protocol_name = "mssql"
        self.config = config or {}
        
        self.versions = {
            "2008": "0A000000",  # SQL Server 2008
            "2012": "11000000",  # SQL Server 2012
            "2014": "12000000",  # SQL Server 2014
            "2016": "13000000",  # SQL Server 2016
            "2017": "14000000",  # SQL Server 2017
            "2019": "15000000",  # SQL Server 2019
            "2022": "16000000",  # SQL Server 2022
        }
        self.version = self.versions.get(self.config.get("version", "2012"), "11000000")
        self.hostname = self.config.get("hostname", "server-01")

    def connection_made(self, transport):
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

    def data_received(self, data):
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
        
        if data[0] == 0x12: #
            self.pre_login(data)
        elif data[0] == 0x10:
            self.login(data)

    def pre_login(self, data):
        self.transport.write(
            bytes.fromhex(
                "0401002500000100000015000601001b000102"
                f"001c000103001d0000ff{self.version}00000200"
            )
        )

    def login(self, data):
        credentials = self.extract_credentials(data)
        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra=credentials)
        
        auth_response = self.generate_login_error("Login failed for user '" + credentials['username'] + "'.", 18456)
        self.transport.write(auth_response)

    def decrypt_mssql_password(self, encrypted):
        # The decryption is done with a simple XOR and nibble swap
        # https://nvd.nist.gov/vuln/detail/CVE-2002-1872
            
        def decrypt_byte(b):
            # XOR byte with A5
            decrypted_byte = b ^ 0xA5
            # Swap nibbles
            return ((decrypted_byte & 0x0F) << 4) | ((decrypted_byte & 0xF0) >> 4)
        
        decrypted_bytes = bytearray(decrypt_byte(b) for b in encrypted)
        return decrypted_bytes.decode('utf-16-le')

    def extract_credentials(self, data):
        username_start = int.from_bytes(data[48:50], byteorder='little')
        username_length = int.from_bytes(data[50:52], byteorder='little')
        username = data[8 + username_start : 8 + username_start + (username_length * 2)].decode('utf-16-le').strip('\x00')

        password_start = int.from_bytes(data[52:54], byteorder='little')
        password_length = int.from_bytes(data[54:56], byteorder='little')
        password_encoded = data[8 + password_start : 8 + password_start + (password_length * 2)]
        password = self.decrypt_mssql_password(password_encoded)

        return {"username": username, "password": password}

    def generate_login_error(self, error_msg="", error_code=18456):
        if not error_msg:
            error_msg = "Login failed due to invalid credentials."
        
        hostname_len = len(self.hostname).to_bytes(1, 'little')
        error_msg_len = len(error_msg).to_bytes(2, 'little')
        error_code_bytes = error_code.to_bytes(4, 'little')
        
        error_token = (
            error_code_bytes
            + b'\x01\x0e'
            + error_msg_len
            + error_msg.encode('utf-16-le')
            + hostname_len
            + self.hostname.encode('utf-16-le')
            + b'\x00\x01\x00\x00\x00'
        )
        
        done_token = b'\xfd\x02' + b'\x00' * 10
        
        error_token_len = len(error_token).to_bytes(2, 'little')
        packet_data = b'\xaa' + error_token_len + error_token + done_token
        
        total_length = (len(packet_data) + 8).to_bytes(2, 'big')
        header = b'\x04\x01' + total_length + b'\x00\x35\x01\x00'
        
        return header + packet_data

class MssqlHoneypot(BaseHoneypot):
    def __init__(self, config, logger, bindaddr):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: MssqlProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config




from trapster.modules.base import BaseProtocol, BaseHoneypot

import struct
import os

#TODO Switch paquet for clear text password  def auth_switch_request(self, seq_no):

class MysqlProtocol(BaseProtocol):
    '''based on https://dev.mysql.com/doc/internals/en/mysql-packet.html'''

    # auth_plugin can be: 
    # "mysql_native_password" : SHA-1 of password
    # "mysql_clear_password" : clear password
    

    config = {
        "version": "5.6.4-m7-log",
        "auth_plugin": "mysql_native_password",
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "mysql"
        self.protocol_version = b'\x0a'

    def connection_made(self, transport):
        self.transport = transport

        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

        # send initial handshake
        self.transport.write(self.build_packet(self.initial_handshake(), 0))
        
    def data_received(self, data):    
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)

        capability_flag = int.from_bytes(data[0x04:0x08], 'little')

        #CLIENT_PROTOCOL_41
        if not capability_flag & 0x00000200:
            # If client version is less than 4.1
            self.transport.write(b"Client does not support authentication protocol requested by server; consider upgrading MySQL client")
            self.connection_lost(None)
            return

        # ClIENT_SSL
        if capability_flag & 0x00000800:
            # client ask for SSL
            self.connection_lost(None)
            return
     
        max_size = int.from_bytes(data[0x08:0x0C], byteorder='little')
        username_end = data.index(b'\x00', 0x24, max_size - 0x24)
        username = data[0x24:username_end].decode(errors='backslashreplace')

        password_len = data[username_end + 1]
        using_password = "YES" if password_len > 0 else "NO"

        password_end = username_end + 2 + password_len
        password = None
        if password_len > 0:
            # converted to printable hex for loggin
            password = data[username_end + 2:password_end].hex()
        
        extra_details = data[password_end:].decode(errors='backslashreplace')

        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={'username':username, 'password':password, 'details':extra_details})

        local_ip, _ = self.transport.get_extra_info('sockname')

        self.transport.write(
            self.build_packet(
                self.auth_failed("Access denied for user '{}'@'{}' (using password: {})".format(username, local_ip , using_password)),
                2
            )
        )

    def initial_handshake(self):
        # https://dev.mysql.com/doc/dev/mysql-server/latest/page_protocol_connection_phase_packets_protocol_handshake_v10.html
        
        protocol_version = self.protocol_version
        server_version = (self.config['version'] + '\x00').encode()
        connection_id = os.urandom(4) # b'\x56\x0a\x00\x00'
        auth_plugin_data_part_1 = os.urandom(8)
        filler_1 = b'\x00'
        capability_flag_1 = b'\xFF\xF7'
        character_set = b'\x21' #utf8_general_ci
        status_flags = b'\x02\x00'
        capability_flag_2 = b'\xFF\x81'
        auth_plugin_data_len = bytes([21]) # 20 bytes random data + \x00 for mysql_native_password
        zeros = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        auth_plugin_data_part_2 = os.urandom(12) + b'\x00'
        auth_plugin = (self.config['auth_plugin'] + '\x00').encode()

        packet = protocol_version + server_version + connection_id  + auth_plugin_data_part_1 + filler_1 + \
            capability_flag_1 + character_set + status_flags + capability_flag_2 + \
            auth_plugin_data_len + zeros + auth_plugin_data_part_2 + auth_plugin

        return packet

    def auth_failed(self, message:str):
        '''
        int<1>  header  [ff] header of the ERR packet
        int<2>  error_code  error-code
        if capabilities & CLIENT_PROTOCOL_41 {
          string[1]     sql_state_marker    # marker of the SQL State
          string[5]     sql_state   SQL State
        }
        string<EOF>     error_message   human readable error message
        '''

        error_packet = b'\xFF'  # Error Packet ID
        error_code = b'\x15\x04'  # Error code 1045 (0x0415)
        sql_state = bytes("#28000", 'utf-8')
        message = message.encode()

        packet = error_packet + error_code + sql_state + message
        
        return packet


    def build_packet(self, payload, id):
        '''
        int<3>  payload_length  Length of the payload. The number of bytes in the packet beyond the initial 4 bytes that make up the packet header.
        int<1>  sequence_id     Sequence ID
        string<var>     payload     [len=payload_length] payload of the packet
        '''
        payload_length = struct.pack("<I", len(payload) )[:3]
        sequence_id = struct.pack("<B", id)
        packet = payload_length + sequence_id + payload

        return packet


class MysqlHoneypot(BaseHoneypot):

    def __init__(self, config, logger, bindaddr='0.0.0.0'):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: MysqlProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config
        self.handler.config['host'] = bindaddr



from aiohttp import web
from jinja2.sandbox import ImmutableSandboxedEnvironment
from jinja2 import FileSystemLoader, Undefined
import yaml
import random, string, base64, mimetypes, re
from datetime import datetime, timezone
from pathlib import Path

from trapster.modules.base import BaseHoneypot
from trapster.libs.ai.http import HttpAI

class HttpHandler:
    def __init__(self, config, logger):
        self.protocol_name = "http"

        self.logger = logger

        self.NAME = config.get('skin', 'default_apache')
        self.BASIC_AUTH = config.get('basic_auth', False)
        self.USERNAME = config.get('username', None)
        self.PASSWORD = config.get('password', None)

        self.data_folder = Path(__file__).parent.parent / "data" / "http"

    def setup(self):
        try:
            resolved_path = (self.data_folder / self.NAME).resolve()
            if not resolved_path.is_relative_to(self.data_folder):
                raise ValueError(f"Invalid skin name: {self.NAME}")
        except (ValueError, RuntimeError):
            self.NAME = "default_apache"  # Fallback to a default skin

        self.static_folder = self.data_folder / self.NAME / "files"
        self.template_folder = self.data_folder / self.NAME / "templates"
        self.config_file = self.data_folder / self.NAME / "config.yaml"

        with self.config_file.open('r') as file:
            self.http_config = yaml.safe_load(file)
        
        self.env = self.create_jinja_env()

    @staticmethod
    def parse_query_string(query_string):
        # Parse query string into a dictionary
        query_dict = {}
        if query_string:
            for param in query_string.split('&'):
                if '=' in param:
                    key, value = param.split('=', 1)
                    query_dict[key] = value
                else:
                    query_dict[param] = ''
        return query_dict

    async def sanitize_request(self, request):
        if request:
            return { 
                "url": request.url,
                "path": request.path,
                "method": request.method,
                "headers": dict(request.headers),
                "body": await request.text() if request.body_exists else None,
                "remote": request.remote, 
                "cookies": request.cookies,
                "query_string": self.parse_query_string(request.query_string),
                "content_type": request.content_type,
                "host": request.host,
                "secure": request.secure,
                "scheme": request.scheme,
                "path_qs": request.path_qs
            }
        
    @staticmethod
    def random_filter(seed=None, alphabet=string.hexdigits[:-6], length=36):
        # Jinja filter to generate a random string
        if seed is not None:
            random.seed(seed)
        return ''.join(random.choice(alphabet) for _ in range(length))

    def create_jinja_env(self):
        env = ImmutableSandboxedEnvironment(
            loader=FileSystemLoader(self.template_folder),
            autoescape=True
        )
        env.globals.update({
            'random': self.random_filter,
            'get_current_time': lambda: datetime.now(timezone.utc).strftime('%a, %d %b %Y %H:%M:%S GMT'),
        })
        env.undefined = Undefined
        return env

    def get_endpoint_config(self, full_url, method):
        # Parse URL and query parameters
        base_url = full_url.split('?')[0]
        query_string = full_url.split('?')[1] if '?' in full_url else ''
        query_params = {}
        if query_string:
            for param in query_string.split('&'):
                if '=' in param:
                    key, value = param.split('=', 1)
                    query_params[key] = value
                else:
                    query_params[param] = ''

        for endpoint in self.http_config.get('endpoints', []):
            for route, details in endpoint.items():
                # 1. Check base URL match first
                if not re.fullmatch(route, base_url):
                    continue

                # 2. Check method match
                if not isinstance(details, list):
                    details = [details]  # Convert single config to list format

                # Find all configs that match the method
                matching_configs = [d for d in details if d['method'] == method]
                if not matching_configs:
                    continue

                # 3. Handle query parameters
                if not query_params:
                    # URL has no query params - look for config without query rules
                    no_query_config = next((d for d in matching_configs if not d.get('query')), None)
                    if no_query_config:
                        return no_query_config
                else:
                    # URL has query params - try to find matching query rules
                    for config in matching_configs:
                        if not config.get('query'):
                            continue
                        
                        # Check each query parameter rule
                        matches_all = True
                        query_rules = config['query']
                        for param_name, pattern in query_rules.items():
                            if param_name not in query_params or not re.fullmatch(pattern, query_params[param_name]):
                                matches_all = False
                                break
                        if not matches_all:
                            break
                        
                        if matches_all:
                            return config

                    # If no query rules matched but we have a base URL match,
                    # return the first config without query rules
                    no_query_config = next((d for d in matching_configs if not d.get('query')), None)
                    if no_query_config:
                        return no_query_config
        
        return None

    def parse_front_matter(self, content):
        """
        This allows the rendered template to overwrite the status code of the request using:
            ---
            status_code: 500
            ---

        Returns:
            tuple: A dictionary of metadata (e.g., status_code) and the template content.
        """
        if content.startswith('---'):
            parts = content.split('---', 2)
            if len(parts) >= 3:
                header, body = parts[1].strip(), parts[2].strip()
                metadata = {}
                for line in header.splitlines():
                    if ':' in line:
                        key, value = line.split(':', 1)
                        metadata[key.strip()] = value.strip()
                return metadata, body
        return {}, content  # No front matter found, return empty metadata

    async def get_content(self, endpoint_config, request=None):
        if not endpoint_config:
            return "", 200
        
        if 'content' in endpoint_config:
            return endpoint_config['content'], endpoint_config.get('status_code', 200)
        
        elif 'file' in endpoint_config:
            file_path = self.template_folder / endpoint_config['file']
            try:
                if file_path.resolve().relative_to(self.template_folder.resolve()):
                    file_path = file_path.resolve()
                    # Read the template file
                    with file_path.open('r') as file:
                        raw_content = file.read()

                    # Add the request object to the template and render it
                    template = self.env.from_string(raw_content)
                    template.globals['request'] = await self.sanitize_request(request)
                    rendered_output = template.render()

                    # Parse the front matter and template content
                    metadata, template_content = self.parse_front_matter(rendered_output)

                    # Extract the status_code from metadata or default to 200
                    status_code = int(metadata.get('status_code', 200))

                    return template_content, status_code
            except (ValueError, FileNotFoundError):
                pass

        elif 'ai' in endpoint_config:
            # experimental ai response
            ai_agent = HttpAI()
            peer_addr = request.transport.get_extra_info('peername')[0]
            session_id = peer_addr
            result = await ai_agent.make_query("http:"+session_id, request.query_string)
            result = result.replace('```json\n', '').replace('\n```', '') # sometime the AI response is wrapped in ```json
            return result, 200
        
        return "File not found", 404

    async def handle_error(self, request, error_code):
        # Check if errors is defined in config.yaml, otherwise use a error template (like 401.html), otherwise send nothing
        error_config = self.http_config.get('errors', {}).get(str(error_code))
        if error_config:
            content, _ = await self.get_content(error_config)
        else:
            content, _ = await self.get_content({"file": f"{error_code}.html"})

        headers = self.http_config.get('headers', {}).copy()
        headers.update(self.http_config.get('errors', {}).get(str(error_code), {}).get('headers', {}))
        
        if error_code == 401:
            headers['WWW-Authenticate'] = 'Basic realm="Restricted Area"'
        
        await self.log(request, self.logger.QUERY, error_code)
        return web.Response(body=content, content_type='text/html', status=error_code, headers=headers)

    async def handle_request(self, request):
        if self.BASIC_AUTH and not self.check_auth(request):
            return await self.handle_error(request, 401)

        full_url = request.url.path_qs
        method = request.method
        endpoint_config = self.get_endpoint_config(full_url, method)
        headers = {}

        if endpoint_config:
            # Use configured response    
            content, status_code = await self.get_content(endpoint_config, request)
            status_code = endpoint_config.get('status_code', status_code)
            headers = endpoint_config.get('headers', {})
            if method == "POST":
                await self.log(request, self.logger.LOGIN, status_code)
            else:
                await self.log(request, self.logger.QUERY, status_code)

        else:
            content, status_code, headers = await self.handle_static_file(request)
            # only default response is logged

        # Prepare response headers
        response_headers = self.http_config.get('headers', {}).copy()
        response_headers.update(headers)

        # Determine content type, we cannot use both content_type variable and Content-Type in headers
        content_type = response_headers.pop('Content-Type', 'text/html')

        return web.Response(body=content, content_type=content_type, charset='utf-8', 
                            status=status_code, headers=response_headers)

    def check_auth(self, request):
        auth_header = request.headers.get('Authorization')
        if not auth_header or not auth_header.startswith('Basic '):
            return False
        encoded_credentials = auth_header.split(' ', 1)[1]
        username, password = base64.b64decode(encoded_credentials).decode('utf-8').split(':')
        self.logger.log(f"{self.protocol_name}.{self.logger.LOGIN}", request.transport, 
                        extra={"username": username, "password": password})
        return username == self.USERNAME and password == self.PASSWORD

    async def handle_static_file(self, request):
        if request.path.endswith('/'):
            file_path = self.static_folder / request.path.lstrip('/') / 'index.html'
        elif request.method == 'GET':
            file_path = self.static_folder / request.path.lstrip('/')
        else:
            return await self.handle_default(request)

        try:
            if file_path.resolve().relative_to(self.static_folder) and file_path.is_file():
                content = file_path.read_bytes()
                content_type = mimetypes.guess_type(str(file_path))[0] or 'application/octet-stream'
                return content, 200, {'Content-Type': content_type}
        except ValueError:
            return "", 500, {}

        return await self.handle_default(request)

    async def handle_default(self, request):
        default_response = self.http_config.get('default')
        content, _ = await self.get_content(default_response)
        status_code = default_response.get('status_code', 404)
        headers = default_response.get('headers', {})
        await self.log(request, self.logger.QUERY, status_code)
        return content, status_code, headers
    
    async def log(self, request, log_type, status_code, extra=None):
        '''
        Log the request to the logger and extract login and password from the request body, if any
        The POST data is logged as hex string in the data field
        The rest is processed and logged in the extra field
        '''

        all_extra = {
            "method": request.method,
            "target": request.path_qs,
            "version": f"HTTP/{request.version.major}.{request.version.minor}",
            "headers": dict(request.headers),
            "status_code": status_code,
        }
        all_extra.update(extra or {})

        data = ''
        if request.body_exists:
            data = await request.read()
            if data:
                form_data = data.decode('utf-8')
                
                query_direct = self.parse_query_string(form_data)
                for key, value in query_direct.items():
                    if key in ['login', 'username','account', 'user%5Blogin%5D', 'j_username']:
                        all_extra['username'] = value
                    elif key in ['password', 'credential', 'passwd', 'user%5Bpassword%5D', 'j_password']:
                        all_extra['password'] = value
                
                # Handle XML/SOAP data
                if '<Envelope' in form_data and '</Envelope>' in form_data:
                    if '<userName>' in form_data and '</userName>' in form_data:
                        username_start = form_data.find('<userName>') + len('<userName>')
                        username_end = form_data.find('</userName>')
                        all_extra['login'] = form_data[username_start:username_end]
                        
                    if '<password>' in form_data and '</password>' in form_data:
                        password_start = form_data.find('<password>') + len('<password>')
                        password_end = form_data.find('</password>')
                        all_extra['password'] = form_data[password_start:password_end]
        
        self.logger.log(f"{self.protocol_name}.{log_type}", request.transport, data=data ,extra=all_extra)


class HttpHoneypot(BaseHoneypot):
    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.port = config['port']
        self.handler = HttpHandler(config=config, logger=logger)

    async def start(self):
        self.handler.setup()
        app = web.Application()
        app.add_routes([web.route('*', '/{path:.*}', self.handler.handle_request)])
        runner = web.AppRunner(app, access_log=None, handle_signals=True)
        await runner.setup()
        self.site = web.TCPSite(runner, self.bindaddr, self.port)
        await self.site.start()

    async def stop(self):
        if hasattr(self, 'site'):
            await self.site.stop()



from setuptools import setup, find_packages
import codecs
import os

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r', encoding='utf-8') as fp:
        return fp.read()

def get_version(rel_path):
    """
    Reading the package version dynamically.
    https://packaging.python.org/en/latest/guides/single-sourcing-package-version/
    """
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

requirements = read('requirements.txt').splitlines()

setup(
    name='trapster',
    version=get_version("trapster/__init__.py"),
    install_requires=requirements,
    url='https://trapster.cloud/',
    author='0xBallpoint',
    author_email='contact@ballpoint.fr',
    description='Trapster Daemon',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license="AGPL3",
    packages=find_packages(include=['trapster', 'trapster.*', 'trapsterd']),
    include_package_data=True,
    package_data={
        '': ['requirements.txt'],
    },
    platforms=['linux'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Framework :: AsyncIO",
        "Topic :: System :: Networking",
        "Topic :: Security",
        "Topic :: System :: Networking :: Monitoring",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)"
    ],
    keywords=["trapster", "honeypot", "ballpoint", "deceptive", "security", "network"],
    entry_points={
        'console_scripts': [
             'trapster=trapster.trapster:main',
        ],
    },
)



import logging
import httpx
from abc import ABC, abstractmethod

from .redis_manager import RedisManager

class HandlerAI(ABC):
    def __init__(self, 
                 api_url="https://llama-3-1-70b-instruct.endpoints.kepler.ai.cloud.ovh.net/api/openai_compat/v1/chat/completions", 
                 api_key=None, 
                 headers=None):
        
        self.MAX_HISTORY_MESSAGES = 50
        self.url = api_url
        self.api_key = api_key
        self.headers = headers or {
            "Content-Type": "application/json",
            #"Authorization": f"Bearer {self.api_key}"
        }
        self.redis_manager = RedisManager()
        self.initial_messages = self._get_initial_messages()

    @abstractmethod
    def _get_initial_messages(self) -> list:
        """
        Each AI handler should define its own initial messages
        It should contains the system prompt and the initial user message
        """
        pass

    async def make_query(self, session_id: str, prompt: str) -> str:
        """
        Make a query to the AI API
        A session is a unique identifier for a user's interaction with the AI (IP, username, etc.)
        """

        # Connect to Redis
        await self.redis_manager.connect()

        # Get existing messages for this session
        messages = await self.redis_manager.get_history(session_id)
        messages = self.initial_messages + messages[:self.MAX_HISTORY_MESSAGES]

        # Add new message
        new_message = {
            "content": prompt,
            "name": "user",
            "role": "user"
        }

        # Check cache first
        cached_response = await self.redis_manager.get_cache(prompt)
        if cached_response is not None:
            # Add both the user message and cached response to history
            await self.redis_manager.add_to_history(session_id, new_message)
            await self.redis_manager.add_to_history(session_id, {
                "content": cached_response,
                "name": "assistant",
                "role": "assistant"
            })
            return cached_response
        
        # If not in cache, proceed with API call
        await self.redis_manager.add_to_history(session_id, new_message)

        payload = {
            "max_tokens": 3000,
            "messages": messages + [new_message],
            "model": "Meta-Llama-3_1-70B-Instruct",
            "temperature": 0,
        }

        try:
            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.post(self.url, json=payload, headers=self.headers)
                logging.debug(response.text)
        except httpx.RequestError as e:
            logging.error(f"RequestError: {e}")
            return ""

        if response.status_code == 200:
            response_data = response.json()    
            choices = response_data["choices"]
            
            for choice in choices:
                result = choice["message"]["content"]
                
                # Save assistant's response to history
                await self.redis_manager.add_to_history(session_id, {
                    "content": result,
                    "name": "assistant",
                    "role": "assistant"
                })
                
                # Save to cache
                await self.redis_manager.set_cache(prompt, result)
                
                return result
        else:
            logging.error("Error:", response.text)
            return ""